﻿using JewelryStoreContracts.BindingModels;
using JewelryStoreContracts.BusinessLogicsContracts;
using JewelryStoreContracts.SearchModels;
using JewelryStoreContracts.StoragesContracts;
using JewelryStoreContracts.ViewModels;
using JewerlyStoreDataModels.Enums;
using JewerlyStoreDataModels.Models;
using Microsoft.Extensions.Logging;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace JewelryStoreBusinessLogic.BusinessLogics
{
    public class OrderLogic : IOrderLogic
    {
        private readonly ILogger _logger;
        private readonly IOrderStorage _orderStorage;
        private readonly IShopStorage _shopStorage;
        private readonly IShopLogic _shopLogic;
        private readonly IJewelStorage _jewelStorage;

        public OrderLogic(ILogger<OrderLogic> logger, IOrderStorage orderStorage, IShopStorage shopStorage, IShopLogic shopLogic, IJewelStorage jewelStorage)
        {
            _logger = logger;
            _orderStorage = orderStorage;
            _shopStorage = shopStorage;
            _shopLogic = shopLogic;
            _jewelStorage = jewelStorage;
        }

        public bool CreateOrder(OrderBindingModel model)
        {
            CheckModel(model);
            if (model.Status != OrderStatus.Неизвестен) return false;
            model.Status = OrderStatus.Принят;
            if (_orderStorage.Insert(model) == null)
            {
                _logger.LogWarning("Insert operation failed");
                return false;
            }
            return true;
        }
        public List<OrderViewModel>? ReadList(OrderSearchModel? model)
        {
            _logger.LogInformation("ReadList. Id:{Id}", model?.Id);
            var list = model == null ? _orderStorage.GetFullList() : _orderStorage.GetFilteredList(model);
            if (list == null)
            {
                _logger.LogWarning("ReadList return null list");
                return null;
            }
            _logger.LogInformation("ReadList. Count:{Count}", list.Count);
            return list;
        }
        public dynamic ElementForOrder(OrderBindingModel model) 
        {
            CheckModel(model, false);
            var element = _orderStorage.GetElement(new OrderSearchModel
            {
                Id = model.Id
            });
            if (element == null)
            {
                _logger.LogWarning("Read operation failed");
                return false;
            }
            return element;
        }

        public bool DeliveryOrder(OrderBindingModel model)
        {
            var element = ElementForOrder(model);
            if (element.Status != OrderStatus.Готов)
            {
                _logger.LogWarning("Status change operation failed");
                throw new InvalidOperationException("Заказ должен быть переведен в статус готовности перед выдачей!");
            }
            model.Status = OrderStatus.Выдан;
            model.DateImplement = DateTime.Now;
            _orderStorage.Update(model);
            return true;
        }

        public bool FinishOrder(OrderBindingModel model)
        {
            var element = ElementForOrder(model);
            if (element.Status != OrderStatus.Выполняется)
            {
                _logger.LogWarning("Status change operation failed");
                throw new InvalidOperationException("Заказ должен быть переведен в статус выполнения перед готовностью!");
            }           
            model.Status = OrderStatus.Готов;
            var jewel = _jewelStorage.GetElement(new JewelSearchModel() { Id = model.JewelId });
            if (jewel == null)
            {
                _logger.LogWarning("Status change error. Jewel not found");
                return false;
            }
            if (!CheckSupply(jewel, model.Count))
            {
                _logger.LogWarning("Status change error. Shop doesnt have jeweles");
                return false;
            }
            _orderStorage.Update(model);
            return true;
        }

        public bool TakeOrderInWork(OrderBindingModel model)
        {
            var element = ElementForOrder(model);
            if (element.Status != OrderStatus.Принят)
            {
                _logger.LogWarning("Status change operation failed");
                throw new InvalidOperationException("Заказ должен быть переведен в статус принятого перед его выполнением!");
            }
            model.Status = OrderStatus.Выполняется;
            _orderStorage.Update(model);
            return true;
        }

        private void CheckModel(OrderBindingModel model, bool withParams = true)
        {
            if (model == null)
            {
                throw new ArgumentNullException(nameof(model));
            }
            if (!withParams)
            {
                return;
            }
            if (model.Sum <= 0)
            {
                throw new ArgumentNullException("Сумма заказа должна быть больше 0", nameof(model.Sum));
            }
            if (model.Count <= 0)
            {
                throw new ArgumentNullException("Количество изделий должно быть больше 0", nameof(model.Count));
            }
            _logger.LogInformation("Order. Sum:{Sum}. Id:{Id}", model.Sum, model.Id);
            var element = _orderStorage.GetElement(new OrderSearchModel
            {
                Id = model.Id
            });
            if (element != null && element.Id != model.Id)
            {
                throw new InvalidOperationException("Заказ с таким ID уже существует");
            }
        }

        public bool CheckSupply(IJewelModel jewel, int count)
        {
            if (count <= 0)
            {
                _logger.LogWarning("Check supply operation error. Jewel count < 0");
                return false;
            }
            int Capacity = _shopStorage.GetFullList().Select(x => x.MaxCountJewels).Sum();
            int currentCount = _shopStorage.GetFullList().Select(x => x.ShopJewels.Select(y => y.Value.Item2).Sum()).Sum();
            int freeSpace = Capacity - currentCount;

            if (freeSpace < count)
            {
                _logger.LogWarning("Check supply error. No place for new Jewels");
                throw new Exception("Нет места для новых драгоценностей");
            }
            foreach (var shop in _shopStorage.GetFullList())
            {
                freeSpace = shop.MaxCountJewels;
                foreach (var doc in shop.ShopJewels)
                {
                    freeSpace -= doc.Value.Item2;
                }
                if (freeSpace == 0)
                {
                    continue;
                }
                if (freeSpace >= count)
                {
                    if (_shopLogic.AddJewel(new()
                    {
                        Id = shop.Id
                    }, jewel, count))
                    {
                        count = 0;
                    }
                    else
                    {
                        _logger.LogWarning("Supply error");
                        return false;
                    }
                }
                else
                {
                    if (_shopLogic.AddJewel(new() { Id = shop.Id }, jewel, freeSpace))
                        count -= freeSpace;
                    else
                    {
                        _logger.LogWarning("Supply error");
                        return false;
                    }
                }
                if (count <= 0)
                {
                    return true;
                }
            }
            return false;
        }
    }
}
